<!--	
	Library of routines to perform low-level scheduling, analogous to old Post_Inline;
	
	Implements:
		[Event_Schedule]

	Variables:
		__Schedule_Status__ is Active/Sleep/Pause/Abort
			Active - Check schedule queue within one minute, sleep until next event 
			Sleep - Sleep until indicated time
			Pause - Pause indefinitely
			Abort - Kill the session watcher
		__Schedule_Sleep__ is date/time at which to wake up

	History:
		10/30/2003	JS	- added special handling of delays that are multiples of 1440
		10/30/2003	JS	- added async execution of events
		10/30/2003	JS	- added custom tag _include_async
		10/30/2003	JS	- removed -NoReply from include_url since that 
							still waits for the page so it doesn't help much
	    10/1/2003   FMS - LP7 syntax update, privileged tag, new status variable, silent sleep.
	    08/13/2002  FMS - LP6 syntax update, optimized SQL statements, added logging.
	    05/30/2002  FMS - Updated with new syntax, optimized SQL statements
		12/07/2001	GLH - Changed max expire date from 3000 to 2021 to prevent math overflow
		09/07/2001	GLH - Changed to _LLEvent_Schedule; to get around security issues
		06/18/2001	GLH - Added schedule pause logic
		04/11/2001	GLH - Initial Development
-->

<!-- 
	Event_Schedule:
		Queues a schedule item for later execution.  Roughly equivalent to old Post_Inline;

	Inputs:
		-Start=(Date),			//	defaults to today
		-End=(Date),			//	defaults to never
		-URL=(ExecuteURL),		//	URL to execute (simulate a browser hit) (Required)
	    -Repeat=(true/false),	//	Should this repeat? Default to true if -Delay passed
	    -Restart=(true/false),	//	Continue executing after restart.  Defaults to true
	    -Delay=(#Minutes),
	    -Username,				//	optional username for authentication
	    -Password;				//	optional password for authentication

	Outputs:
		None

	Side-Effects:
		Appends a record to site._Schedule that contains all info required to execute the item
-->

<?LassoScript

// moved block inside async

Define_Tag: 'Event_Schedule', -Priority='replace', -Privileged,
		-Required='url',
		-Optional='start',
		-Optional='end',
		-Optional='delay',
		-Optional='repeat',
		-Optional='restart',
		-Optional='username',
		-Optional='password';

	// Create an array to store the constructed event
	local: 'event_array' = (array);

	// Insert the event
	local: 'l_url' = (string: (local: 'url'));
	#l_url->trim;
	#event_array->(insert: 'executeURL' = (local: 'l_url'));

	log_warning: 'Schedule: Queue URL <' + (local: 'l_url') + '>';

	// If no start date defined, then default to today
	if: ((local: 'start') == '');
		#event_array->(insert: 'date_time_start' = (Date)->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date)->(Format: '%Q %T'));
	else;
		#event_array->(insert: 'date_time_start' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
		#event_array->(insert: 'next_execute' = (Date: (string: (local: 'start')))->(Format: '%Q %T'));
	/If;

	// If no end date defined, then default to never
	if: ((local: 'end') == '');
		#event_array->(insert: 'date_time_end' = '3001-01-01 00:00:00');
	else;
		#event_array->(insert: 'date_time_end' = (Date: (string: (local: 'end')))->(Format: '%Q %T'));
	/if;

	// If no delay defined, then default to 0
	if: ((local: 'delay') == '');
		local: 'l_delay' = 0;
	else: ((integer: (local: 'delay')) < 0);
		local: 'l_delay' = 0;
	else;
		local: 'l_delay' = (integer: (local: 'delay'));
	/if;
	#event_array->(insert: 'delay'=#l_delay);

	// If no repeat defined, then default to true iff a delay is passed
	if: ((local: 'repeat') == '');
		if: (#l_delay > 0);
			#event_array->(insert: 'repeat'='1');
		else;
			#event_array->(insert: 'repeat'='0');
		/if;
	else: (#l_delay == 0);
		#event_array->(insert: 'repeat'='0');
	else;
		#event_array->(insert: 'repeat'=(integer: (boolean: (local: 'repeat'))));
	/if;

	// If no restart defined, then default to true
	if: ((local: 'restart') == '');
		#event_array->(insert: 'restart'='1');
	else;
		#event_array->(insert: 'restart'=(integer: (boolean: (local: 'restart'))));
	/if;

	// If no username defined, then default to ''
	#event_array->(insert: 'user'=(string: (local: 'username')));

	// If no password defined, then default to ''
	#event_array->(insert: 'pass'=(string: (local: 'password')));

	// Add the event to the database
	inline: -database='site', -table='_Schedule', -add, #event_array;
		if: (error_currenterror) != (error_noerror);
			log_warning: 'Schedule: ' + (error_currenterror) + ' (' + (error_currenterror: -errorcode) + ')';
		/if;
	/inline;
	
	// After adding an event, check the event queue immediately
	Global: '__Schedule_Status__' = 'Active';

/Define_Tag;

?>

<!--
	Schedule Code

	Actions:
		Creates new _Schedule table if it doesn't exist already
		Changes the type of the executeURL column if it is not TEXT
		Creates global __Schedule_Status__ variable
		Creates global __Schedule_Sleep__ variable
		Deletes events that are not set to restart
		Starts watcher and loops waiting for events to be scheduled

	Note:
		This code is only designed to be called once from LassoStartup
		It is formatted as a compound expression so Local variables
		can be used.
		
-->

<?LassoScript
	{
		Log_Warning: 'Schedule: Initializing Event Queue';
	
		var: '__Schedule_Status__' = 'Active';
		var: '__Schedule_Sleep__' = (Date);
		
		inline: -database='lasso_internal', -table='global_prefs', -search, -op='eq', 'store_key'='lasso_schedule_pause';
			if: (field: 'data') == 'Y';
				var: '__Schedule_Status__'='Pause';
			else;
				var: '__Schedule_Status__'='Active';
			/if;
		/inline;
	
		// Check if the table exists
		Local: 'table_exists' = false;
		local: 'table_fields' = (map);
		Inline: -Database='site', -SQL='SHOW COLUMNS FROM _Schedule';
			if: (found_count) > 0;
				#table_exists = true;
				records;
					#table_fields->(insert: (field: 'field')=(field: 'type'));
				/records;
			/if;
		/inline;
	
		if: #table_exists == true;
			// Alter table if it does not contain new fields
			If: !(#table_fields->(find: 'executeURL') >> 'TEXT');
				Inline: -Database='site',
					-SQL='ALTER TABLE _Schedule CHANGE COLUMN executeURL executeURL TEXT';
				/Inline;
			/if;
		else;
			// Create table if it doesn't exist already
			Inline: -Database='site',
				-SQL='CREATE TABLE IF NOT EXISTS _Schedule ' +
				'(' +
					'id INT PRIMARY KEY AUTO_INCREMENT,' +
					'date_time_start DATETIME,' +
					'date_time_end DATETIME,' +
					'next_execute DATETIME,' +
					'executeURL TEXT,' +
					'repeat TINYINT,' +
					'restart TINYINT,' +
					'delay INT,' +
					'user VARCHAR(32),' +
					'pass VARCHAR(32),' +
					'INDEX (date_time_end),' +
					'INDEX (next_execute)' +
				')';
			/Inline;
		/if;
	
		// Delete all records that are marked as non-restart or are past end date
		Inline: -Database='site',
			-SQL='DELETE FROM _Schedule WHERE restart=0 OR date_time_end <= NOW() OR next_execute = \'\'';
		/Inline;
	
		// Push the next_execute date to the next even multiple
		// This prevents events from ever executing twice rapidly
		Inline: -Database='site', -SQL='SELECT * FROM _Schedule WHERE date_time_end >= NOW() AND next_execute <= NOW() ORDER BY next_execute';
			Records;
				local: 'next_execute' = (Date: (Field: 'next_execute'), -DateFormat='%Y-%m-%d %H:%M:%S');
				local: 'date_fix' = (Date)->(Difference: (String: (Date)), -MinutesBetween);
				local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
				
				local: 'delay' = (integer: (field: 'delay'));
	
				If: (#delay > 0);
					local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
					#next_execute->(add: -minute=#next_difference);
					
					Inline:  -Database='site', -SQL='UPDATE _Schedule SET next_execute=\"' + (Date_Format: #next_execute, -DateFormat='%Y-%m-%d %H:%M:%S') + '\" WHERE id = ' + (Field: 'id');
					/Inline;
				/If;
			/Records;
		/Inline;
		
	}->(Run);

	// The rest of the script is run asynchronously

	{
		// 
		// Start Schedule Watcher
		//

		// added by Johan Slve 10/30/2003
		define_tag: '_include_async', -async,
			-required='url',
			-optional='username',
			-optional='password',
			-priority='replace';
			// runs include_url in a separate thread
			if: (local_defined: 'username');
				include_url: #url,
					-username=(local: 'username'),
					-password=(local: 'password');
			else;
				include_url: #url;
			/if;
		/define_tag;

		Log_Warning: 'Schedule: Starting Event Queue';
		
		// 30 Second Delay at Startup
		Sleep: 30000; 
		
		While: ((Global: '__Schedule_Status__') != 'Abort');
			Protect;
				If: ((Global: '__Schedule_Status__') == 'Pause') || ((Global: '__Schedule_Status__') == 'Stop');
					local: 'execute' = false;
				else: ((Global: '__Schedule_Status__') == 'Sleep');
					if: (Global: '__Schedule_Sleep__')->(isa: 'date') && ((Global: '__Schedule_Sleep__') <= Date);
						local: 'execute' = true;
					else;
						local: 'execute' = false;
					/if;
				else;
					local: 'execute' = true;
				/if;
				
				If: (#execute == true);
				
					//
					// Check Event Queue
					//
				
					Log_Warning: 'Schedule: Checking Event Queue';
						
					// Delete all records that are past end date.
					Inline: -Database='site', -SQL='DELETE FROM _Schedule WHERE date_time_end <= NOW()';
					/Inline;
				
					// Set the sleep time to one hour
					Global: '__Schedule_Sleep__' = (Date);
					(Global: '__Schedule_Sleep__')->(Add: -Hour=1);
				
					// Execute all records that are not past end date and which have next_execute earlier than now.
					Inline: -Database='site', -SQL='SELECT * FROM _Schedule WHERE date_time_end >= NOW() AND next_execute <= NOW() ORDER BY next_execute';
						Records;
							If: ((Global: '__Schedule_Status__') == 'Pause') || ((Global: '__Schedule_Status__') == 'Stop');
								Loop_Abort;
							/If;
							
							local: 'executeURL' = (field: 'executeURL');
							#executeURL->(trim);
				
							log_warning: 'Schedule: Executing URL <' + (local: 'executeURL') + '>';
				
							Protect;
								//  Now execute the URL (do not wait for results)
								If: (Field: 'user') == '';
									_Include_async: -url=(local: 'executeURL');
								Else;
									_Include_async: -url=(local: 'executeURL'), -Username=(Field: 'user'), -Password=(Field: 'pass');
								/If;
							/Protect;
							
							local: 'delay' = (integer: (field: 'delay'));
							local: 'repeat' = (integer: (field: 'delay'));
				
							// Delete the record if it's a one-shot
							If: (#repeat == 0) || (#delay <= 0);
								log_warning: 'Schedule: Deleting One-Time Event';
								Inline:  -Database='site', -SQL='DELETE FROM _Schedule WHERE id = ' + (Field: 'id'); 
								/Inline;
							Else;
								local: 'next_execute' = (Date: (Field: 'next_execute'), -DateFormat='%Q %T');
								if: #delay % 1440 == 0;
									// handle daily, weekly etc repeats specially to make them repeat on the same time of day,
									// regardless of clock changes like daylight savings switch
									// Added by Johan Slve 10/30/2003
									#next_difference=0;
									// split #next_execute into time and date (use neutral time for the date)
									local: 'next_execute_time'= #next_execute -> format: '%T';
									local: 'next_execute_date'= (date: (#next_execute -> format: '%Q 12:00:00'));
									// add the delay to get the right date (we will disregard the time from this operation)
									#next_execute_date->(add: -minute=#delay);
									// restore the previous time value
									#next_execute = date: (#next_execute_date -> format: ('%Q ' + #next_execute_time));
								else;
									// normal rescheduling
									local: 'date_fix' = (Date)->(Difference: (String: (Date)), -MinutesBetween);
									local: 'next_difference' = #next_execute->(Difference: (Date), -MinutesBetween) + #date_fix;
									local: 'next_difference' = #next_difference + #delay - (#next_difference % #delay);
									#next_execute->(add: -minute=#next_difference);
								/if;
								log_warning: 'Schedule: Rescheduling to ' + (local: 'next_execute');
								Inline:  -Database='site', -SQL='UPDATE _Schedule SET next_execute=\"' + (Date_Format: #next_execute, -DateFormat='%Q %T') + '\" WHERE id = ' + (Field: 'id');
								/Inline;
								If: (global: '__Schedule_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Schedule_Sleep__'));
									global: '__Schedule_Sleep__' = #next_execute;
								/if;
							/If;
							
							// 5 Second Delay Between Events
							If: (Loop_Count < Found_Count);
								Sleep: 5000; 
							/If;
							
						/Records;
					/Inline;

					if: ((Global: '__Schedule_Status__') != 'Pause') && ((Global: '__Schedule_Status__') != 'Stop');
						Inline: -Database='site', -SQL='SELECT next_execute FROM _Schedule ORDER BY next_execute LIMIT 1';
							If: (Found_Count > 0);
								local: 'next_execute' = (Date: (Field: 'next_execute'), -DateFormat='%Q %T');
								If: (global: '__Schedule_Sleep__')->(isa: 'date') && (#next_execute < (global: '__Schedule_Sleep__'));
									global: '__Schedule_Sleep__' = #next_execute;
								/if;
							/if;
						/inline;
						
						If: (global: '__Schedule_Sleep__')->(isa: 'date') && ((date_add: (date), -minute=2) < (global: '__Schedule_Sleep__'));
							global: '__Schedule_Status__' = 'Sleep';
							Log_Warning: 'Schedule: Sleeping Until ' + (global: '__Schedule_Sleep__');
						else;
							global: '__Schedule_Status__' = 'Active';
						/if;
					/if;

				/If;
				
			/Protect;
			
			// 60 Second Delay Between Active Checks
			Sleep: 60000
			; 
		/While;
		
		Log_Warning: 'Schedule: Stopping Event Queue';	

	}->(AsAsync); 
?>
